import json
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple

import pandas as pd
import requests
from dateutil import tz
from dateutil.relativedelta import relativedelta

API_KEY = "API KEY HERE"
BASE_URL = "BASE URL HERE"
HEADERS = {
    "accept": "application/json",
    "x-api-key": API_KEY,
}

WINDSPEED = "WindSpd_Avg"
POWER = "ActivePower_Min"
FIELDS = [WINDSPEED, POWER]

# whether we are running a local backtest or running the deployed
# rule in production
RUNNING_LOCAL_BACKTEST = False
BACKTEST_FILE_PATH = "backtest_results.csv"


def get_asset_timezone_info(asset_id: int) -> Tuple[str, str]:
    """Get the timezone information for an asset.

    Parameters
    ----------
    asset_name
        name of the asset to get information about

    Returns
    -------
    timezone_preference, timezone
        timezone preference (either "UTC" or "asset") and local
        timezone for the given asset

    """
    url = BASE_URL + "assets"
    response = requests.get(url, headers=HEADERS)
    df = pd.DataFrame(response.json()["data"])

    current_asset = df[df["asset_id"] == asset_id]
    timezone_preference = current_asset["timezone_preference"].item()
    timezone = current_asset["timezone"].item()

    return timezone_preference, timezone


def convert_to_asset_time(timestamp: datetime, asset_timezone: str) -> datetime:
    """Convert the given timestamp in UTC time to asset time.

    Parameters
    ----------
    timestamp
        timestamp to convert, in UTC time

    asset_timezone
        timezone string representing local timezone of the asset

    Returns
    -------
    datetime
        datetime object representing timestamp in asset time

    """
    from_zone = tz.gettz("UTC")
    to_zone = tz.gettz(asset_timezone)
    timestamp = timestamp.replace(tzinfo=from_zone)
    return timestamp.astimezone(to_zone)


def get_all_devices_for_asset(asset_id: int) -> Dict[str, Any]:
    """Get JSON containing information on all devices for the given asset_id.

    Parameters
    ----------
    asset_id
        ID for the asset for which we want all devices

    Returns
    -------
    raw_device_info
        JSON with information on all devices for the given asset

    """
    url = BASE_URL + "devices"
    payload = {"asset_id": asset_id}
    response = requests.get(url, headers=HEADERS, params=payload)
    raw_device_info = response.json()["data"]
    return raw_device_info


def get_device_names_and_ids(raw_device_info: Dict[str, Any]) -> Dict[str, int]:
    """Get dictionary with device name as key and device ID as value,
    excluding met and site devices.

    Parameters
    ----------
    raw_device_info
        JSON with information on all devices for an asset

    Returns
    -------
    devices
        dictionary with device name as key and device ID as value, excluding
        met and site devices

    """
    devices = {
        device["name"]: device["device_id"]
        for device in raw_device_info
        if "met" not in device["name"].lower() and "site" not in device["name"].lower()
    }
    return devices


def get_data(
    asset_id: int,
    device_ids: List[int],
    num_timestamps: int,
    timezone_preference: str,
    timezone: str,
    fields: List[str],
    start_date: Optional[datetime] = None,
    end_date: Optional[datetime] = None,
) -> pd.DataFrame:
    """Get data for the given asset and devices.

    Parameters
    ----------
    asset_id
        asset ID representing which asset to get data for

    device_ids
        list of device ids to get data for

    num_timestamps
        number of timestamps to get data for

    timezone_preference
        either "asset" or "utc" -- represents timezone preference for
        the given asset

    timezone
        timezone string representing timezone of the given asset;
        if timezone_preference is "asset", the times used in the query are
        converted to this timezone

    fields
        list of fields to query data for

    start_date
        optional start date to pass in if running a backtest

    end_date
        optional end date to pass in if running a backtest

    Returns
    -------
    data
        DataFrame with data for the given asset and devices -- contains the
        following columns: device, timestamp, timestamp_tz, fields requested

    """
    url = BASE_URL + "data/query"

    # when using the data API, if the timezone preference is "asset", the
    # query must have start_date and end_date in the asset timezone
    # if the timezone preference is "UTC", then start_date and end_date
    # must be provided in UTC time
    if start_date is None and end_date is None:
        if timezone_preference == "asset":
            end_date = convert_to_asset_time(datetime.now(tz=tz.gettz("UTC")), timezone)
        else:
            end_date = datetime.now(tz=tz.gettz("UTC"))
        start_date = end_date - relativedelta(minutes=(10 * num_timestamps))

    payload = {
        "asset_id": asset_id,
        "start_date": start_date.strftime("%Y-%m-%dT%H:%M:%S"),
        "end_date": end_date.strftime("%Y-%m-%dT%H:%M:%S"),
        "fields": fields,
        "device_id": ",".join(map(str, device_ids)),
    }

    response = requests.get(url, headers=HEADERS, params=payload)
    data = pd.DataFrame()
    if "data" in response.json():
        json_data = response.json()["data"]
        data = pd.DataFrame(json_data)
        if not data.empty:
            data["timestamp"] = pd.to_datetime(data["timestamp"])
            data["timestamp_tz"] = pd.to_datetime(data["timestamp_tz"])
            data = data.sort_values(by="timestamp")

    return data


def check_data(data: pd.DataFrame, fields: List[str]) -> bool:
    """Check that the data we pulled is not empty and has all required columns.

    Parameters
    ----------
    data
        DataFrame from the data query

    fields
        list of expected fields that the data should have

    Returns
    -------
    bool
        False if the data is empty or does not have all required columns,
        True otherwise

    """
    if data.empty or not all(
        [
            field in data.columns
            for field in fields + ["timestamp", "timestamp_tz", "device"]
        ]
    ):
        return False
    return True


def get_notification(
    asset_id: int, rule_subject: str, device_id: Optional[int]
) -> Optional[int]:
    """Check if an active notification exists for the given parameters and return
    the notification ID if one does exist.

    Parameters
    ----------
    asset_id
        ID of the asset we want a notification for

    rule_subject
        subject of the notification

    device_id
        device id to get notification for

    Returns
    -------
    notification_id
        ID of an active notification corresponding to the given parameters,
        or None if no active notification exists

    """
    notification_id = None
    url = BASE_URL + "notifications"

    payload = {
        "asset_id": asset_id,
        "status": "active",
        "subject": rule_subject,
        "impacted_device_id": device_id,
    }

    response = requests.get(url, headers=HEADERS, params=payload)
    response = response.json()
    if "data" in response and len(response["data"]) > 0:
        notification_id = response["data"][0]["notification_id"]

    return notification_id


def run_rule(
    data: pd.DataFrame,
    asset_id: int,
    rule_id: int,
    min_timestamps: int,
    devices: Dict[str, int],
    windspeed_threshold: float,
    power_threshold: float,
    send_email: List[str],
    send_sms: List[str],
):
    """Runs rule and generates / updates notifications based on the data.

    Parameters
    ----------
    data
        DataFrame with the fields needed to run the rule

    asset_id
        ID of the asset we are running the rule on

    rule_id
        ID of the custom rule being run

    min_timestamps
        minimum number of timestamps where the rule must be satisfied
        before generating a notification

    devices
        Dictionary of devices to run the rule on -- key is the device
        name and value is the corresponding device ID (we need to provide
        device IDs when generating notifications)

    windspeed_threshold
        minimum windspeed threshold used in the rule

    power_threshold
        maximum active power threshold used in the rule

    send_email
        list of emails to be notified by the rule

    send_sms
        list of numbers to be notified by the rule

    """
    rule_subject = "Example rule"
    url = BASE_URL + "notifications"

    data = data[data[WINDSPEED] > windspeed_threshold]
    offline_df = data[data[POWER] < power_threshold]
    online_df = data[data[POWER] >= power_threshold]

    print(f"offline_df: {offline_df.head()}")
    print(f"online_df: {online_df.head()}")

    for d in offline_df["device"].unique():
        device_offline_df = offline_df[offline_df["device"] == d]
        count = device_offline_df.count()["device"]

        if count >= min_timestamps:
            device_offline_df = device_offline_df.sort_values(by="timestamp")

            notification_id = get_notification(
                asset_id=asset_id,
                rule_subject=rule_subject,
                device_id=devices[d],
            )
            if notification_id:
                print(f"UPDATING NOTIFICATION FOR Asset {asset_id}, device {d}")

                message_time = device_offline_df["timestamp_tz"].iloc[-1]

                payload = {
                    "notification_id": notification_id,
                    "notes": (
                        f"Device {d} is still offline as of "
                        f"{message_time.strftime('%Y-%m-%d %H:%M:%S')}."
                    ),
                }
                response = requests.put(url, headers=HEADERS, params=payload)
                print(response.text)

            else:
                print(f"GENERATING NOTIFICATION FOR Asset {asset_id}, device {d}")

                message_end_time = device_offline_df["timestamp_tz"].iloc[0]
                message_start_time = message_end_time - relativedelta(minutes=10)

                plot_start_time = device_offline_df["timestamp_tz"].iloc[
                    0
                ] - relativedelta(days=1)
                plot_end_time = device_offline_df["timestamp_tz"].iloc[
                    0
                ] + relativedelta(days=1)

                additional_data = {
                    "notification_plots": {
                        "x_axis": "TimeStamp",
                        "y_primary": FIELDS,
                        "y_secondary": [],
                        "devices": [{"id": devices[d], "device_name": d}],
                        "start_date": plot_start_time.strftime("%Y-%m-%dT%H:%M:%S"),
                        "end_date": plot_end_time.strftime("%Y-%m-%dT%H:%M:%S"),
                        "chart_options": [
                            {
                                "fill_gaps": False,
                                "shared_tooltip": True,
                                "separate_by_device": False,
                                "separate_by_variable": False,
                            }
                        ],
                    }
                }

                payload = {
                    "notification_rule_id": rule_id,
                    "status": "active",
                    "subject": rule_subject,
                    "message": (
                        f"Device {d} tripped offline between "
                        f"{message_start_time.strftime('%Y-%m-%d %H:%M:%S')} and "
                        f"{message_end_time.strftime('%Y-%m-%d %H:%M:%S')} and "
                        f"is still offline."
                    ),
                    "user_email": "apm@sparkcognition.com",
                    "impacted_device_ids": str(devices[d]),
                    "asset_id": asset_id,
                    "last_active_timestamp": device_offline_df["timestamp"]
                    .iloc[-1]
                    .strftime("%Y-%m-%dT%H:%M:%S"),
                    "additional_data": json.dumps(additional_data),
                    "send_email": ",".join(send_email),
                    "send_sms": ",".join(send_sms),
                }
                response = requests.post(url, headers=HEADERS, json=payload)
                print(response.text)

    for d in online_df["device"].unique():
        notification_id = get_notification(
            asset_id=asset_id,
            rule_subject=rule_subject,
            device_id=devices[d],
        )
        if notification_id:
            device_online_df = online_df[online_df["device"] == d]
            print(f"CLOSED NOTIFICATION FOR Asset {asset_id}, device {d}")

            first_valid_index = device_online_df.first_valid_index()
            latest_close_time = device_online_df.loc[first_valid_index, "timestamp_tz"]
            earliest_close_time = latest_close_time - relativedelta(minutes=10)

            payload = {
                "notification_id": notification_id,
                "status": "closed_resolved",
                "notes": (
                    f"Device {d} came back online between "
                    f"{earliest_close_time.strftime('%Y-%m-%d %H:%M:%S')} and "
                    f"{latest_close_time.strftime('%Y-%m-%d %H:%M:%S')}."
                ),
            }

            response = requests.put(url, headers=HEADERS, params=payload)
            print(response.text)


def run_rule_backtest(
    data: pd.DataFrame,
    windspeed_threshold: float,
    power_threshold: float,
    results_file_path: str,
) -> pd.DataFrame:
    """Run a backtest on the rule and saves a CSV with all
    events where the condition was satisfied.

    Parameters
    ----------
    data
        DataFrame with the fields needed to run the rule

    min_timestamps
        minimum number of timestamps where the rule must be satisfied

    windspeed_threshold
        minimum windspeed threshold used in the rule

    power_threshold
        maximum active power threshold used in the rule

    results_file_path
        path indicating where to save the backtest results

    """
    data = data[data[WINDSPEED] > windspeed_threshold]
    backtest_df = data[data[POWER] < power_threshold]
    backtest_df = backtest_df.sort_values(by=["device", "timestamp"])
    backtest_df.to_csv(results_file_path, index=False)


def handler(event, context):
    """Example event for this custom container:

    {
        "asset_id": 8,
        "rule_id": 1,
        "send_email": ["test1@sparkcognition.com", "test2@sparkcognition.com"],
        "send_sms": ["555-555-5555"],
        "parameter1": 2,
        "parameter2": 10,
        "parameter3": 100,
    }

    Runs the rule: for asset 8, if WindSpeed_Avg > 10 and ActivePower_Min < 100
    for a device for at least 2 consecutive timestamps, generate a notification
    specifying the device is offline (or update anotification that is already open).

    """
    asset_id = event.get("asset_id")
    print(f"asset_id: {asset_id}")

    rule_id = event.get("rule_id")
    send_email = event.get("send_email")
    send_sms = event.get("send_sms")

    min_timestamps = event.get("parameter1")
    windspeed_threshold = event.get("parameter2")
    power_threshold = event.get("parameter3")

    print(f"min_timestamps: {min_timestamps}")
    print(f"windspeed_threshold: {windspeed_threshold}")
    print(f"power_threshold: {power_threshold}")

    timezone_preference, timezone = get_asset_timezone_info(asset_id=asset_id)

    raw_device_info = get_all_devices_for_asset(asset_id=asset_id)
    devices = get_device_names_and_ids(raw_device_info=raw_device_info)

    if RUNNING_LOCAL_BACKTEST:
        # here we configure what times to run on the backtest on
        # in this example, we run it for 2 hours on Jan 1 (in asset time)
        asset_tzinfo = tz.gettz(timezone)
        backtest_start_time = datetime(2023, 1, 1, 0, 0, 0, tzinfo=asset_tzinfo)
        backtest_end_time = datetime(2023, 1, 1, 2, 0, 0, tzinfo=asset_tzinfo)
        data = get_data(
            asset_id=asset_id,
            device_ids=list(devices.values()),
            num_timestamps=min_timestamps,
            timezone_preference=timezone_preference,
            timezone=timezone,
            fields=FIELDS,
            start_date=backtest_start_time,
            end_date=backtest_end_time,
        )

        if check_data(data=data, fields=FIELDS):
            run_rule_backtest(
                data=data,
                windspeed_threshold=windspeed_threshold,
                power_threshold=power_threshold,
                results_file_path=BACKTEST_FILE_PATH,
            )
        else:
            print("Cannot run backtest. There is no data available.")

    else:
        data = get_data(
            asset_id=asset_id,
            device_ids=list(devices.values()),
            num_timestamps=min_timestamps,
            timezone_preference=timezone_preference,
            timezone=timezone,
            fields=FIELDS,
        )

        if check_data(data=data, fields=FIELDS):
            run_rule(
                data=data,
                asset_id=asset_id,
                rule_id=rule_id,
                min_timestamps=min_timestamps,
                devices=devices,
                windspeed_threshold=windspeed_threshold,
                power_threshold=power_threshold,
                send_email=send_email,
                send_sms=send_sms,
            )
        else:
            print("Cannot run rule. There is no data available.")
